@@ -15,6 +15,7 @@
 #include "bmap.h"
 #include "output.h"
 #include "init.h"
+#include "block.h"
 
 void
 fa_agblock(
@@ -69,7 +70,7 @@ fa_attrblock(
 	bmap_ext_t	bm;
 	uint32_t	bno;
 	xfs_fsblock_t	dfsbno;
-	int		nex;
+	xfs_extnum_t	nex;
 
 	bno = (uint32_t)getbitval(obj, bit, bitsz(bno), BVUNSIGNED);
 	if (bno == 0) {
@@ -97,7 +98,7 @@ fa_cfileoffa(
 	bmap_ext_t	bm;
 	xfs_fileoff_t	bno;
 	xfs_fsblock_t	dfsbno;
-	int		nex;
+	xfs_extnum_t	nex;
 
 	bno = (xfs_fileoff_t)getbitval(obj, bit, BMBT_STARTOFF_BITLEN,
 		BVUNSIGNED);
@@ -127,8 +128,8 @@ fa_cfileoffd(
 	bmap_ext_t	*bmp;
 	xfs_fileoff_t	bno;
 	xfs_fsblock_t	dfsbno;
-	int		nb;
-	int		nex;
+	xfs_extnum_t	nb;
+	xfs_extnum_t	nex;
 
 	bno = (xfs_fileoff_t)getbitval(obj, bit, BMBT_STARTOFF_BITLEN,
 		BVUNSIGNED);
@@ -183,7 +184,7 @@ fa_dfiloffa(
 	bmap_ext_t	bm;
 	xfs_fileoff_t	bno;
 	xfs_fsblock_t	dfsbno;
-	int		nex;
+	xfs_extnum_t	nex;
 
 	bno = (xfs_fileoff_t)getbitval(obj, bit, bitsz(bno), BVUNSIGNED);
 	if (bno == NULLFILEOFF) {
@@ -212,8 +213,8 @@ fa_dfiloffd(
 	bmap_ext_t	*bmp;
 	xfs_fileoff_t	bno;
 	xfs_fsblock_t	dfsbno;
-	int		nb;
-	int		nex;
+	xfs_extnum_t	nb;
+	xfs_extnum_t	nex;
 
 	bno = (xfs_fileoff_t)getbitval(obj, bit, bitsz(bno), BVUNSIGNED);
 	if (bno == NULLFILEOFF) {
@@ -266,7 +267,7 @@ fa_dirblock(
 	bmap_ext_t	*bmp;
 	uint32_t	bno;
 	xfs_fsblock_t	dfsbno;
-	int		nex;
+	xfs_extnum_t	nex;
 
 	bno = (uint32_t)getbitval(obj, bit, bitsz(bno), BVUNSIGNED);
 	if (bno == 0) {
@@ -323,7 +324,9 @@ fa_drtbno(
 		dbprintf(_("null block number, cannot set new addr\n"));
 		return;
 	}
-	/* need set_cur to understand rt subvolume */
+
+	set_rt_cur(&typtab[next], xfs_rtb_to_daddr(mp, bno), blkbb,
+			DB_RING_ADD, NULL);
 }
 
 /*ARGSUSED*/
